#ifndef JNISUB_H
#define JNISUB_H
#include <string_view>
#include <utility>
#include <time.h>
#include <string.h>
#include <stdio.h>
#include <jni.h>
#include <filesystem>
#include <vector>
#include "logs.h"
#include "inout.h"
#include "serial.h"
#include "nfcdata.h"
#include "sensoren.h"
#define VISIBLE __attribute__((__visibility__("default")))

extern Sensoren *sensors;
extern jint parsertype;
struct jniHistory;
struct GlucoseValue;
#include "glucose.h"
//typedef std::vector<GlucoseValue*> jniHistory;
struct scandata;
struct jniHistory:std::vector<GlucoseValue*> {
	const GlucoseValue &get(int i) const {return *operator[](i);}
	}; /*How to use a forward declaration in jnisub.h without mentioning vector?*/
struct AlgorithmResults {
    const jniHistory *hist;
    const GlucoseNow *glu;
/*    const int trendArrow;
    const float rateOfChange;
    */
     const int alarm;
    const jboolean isActionable;
    const int esaMinutesToWait;
    const jboolean lsaDetected;
    bool inserted,removed;
    void setremoved(int rem) {
    	removed=rem;
    	}
    void setinserted(int ins) {
    	inserted=ins;
    	}

void showresults(FILE *stream,scandata *dat) const;
    const jniHistory &history() const {return *hist;}
   const GlucoseNow &currentglucose() const {
	return *glu;
	}
   const jboolean getlsaDetected() const {
   	return lsaDetected;
   	}
   int	trend() const {
	return glu->trendArrow;
	}
  float	RateOfChange() const {
    	return  glu->rateOfChange;
	}
~AlgorithmResults();
};

using namespace std;
struct VISIBLE scandata  ;
struct scandata: public nfcdata {
	bool error=false;
//	time_t nutime;
	data_t *data,*info;
  scandata(std::string_view target) ;
  scandata(std::filesystem::path &pad) : scandata(pad.native()) {}
  scandata(time_t tim,const data_t *datain,const data_t *info):  data(data_t::newex(datain)),info(data_t::newex(info)) {
		settime(tim);
		setdata(data->data());
		}
//  scandata(time_t tim,const std::array<const signed char,344> &raw,const std::array<const signed char,6> &info):  data(data_t::newex(raw)),info(data_t::newex(info)) {
  scandata(time_t tim,const std::array<const uint8_t,344> &raw,const std::array<const uint8_t,6> &info):  data(data_t::newex(raw)),info(data_t::newex(info)) {
		settime(tim);
		setdata(data->data());
		}
  scandata( JNIEnv *env,jbyteArray info,jbyteArray dat,time_t tim=time(nullptr));
  ~scandata(); 
	};
template <typename T>
constexpr int alignstart(int len) {
	constexpr int size=alignof(T);
	return ((len+size-1)/size)*size;
	}
enum state:int {
MESS=0,
COMP=1,
ATTE=2,
FAKE=3
};
class multimmap {
const int maxdats;
int freepos;
public:
Mmap<unsigned char> map;
int &datpos(int datnr) {
	return reinterpret_cast<int*>(map.data())[datnr];
	}
template <typename... T> multimmap(int maxdats,T... args):maxdats(maxdats),freepos(datastart()),map(args...) { }
virtual ~multimmap() {}

int datastart() const {
	return maxdats*sizeof(int);
	}
data_t *get(int datnr) {
	if(!map.data())
		return nullptr;
	int pos=datpos(datnr);
	if(pos<datastart())
		return nullptr;
	data_t *dat=reinterpret_cast<data_t*>(map.data()+pos);
	if(dat->len>0)
		return dat;	
	return nullptr;
	}
bool good() const {
	return const_cast<multimmap*>(this)->get(MESS);
	}
data_t * alloc(int len) {
	int was=freepos;
	data_t *dat=reinterpret_cast<data_t*>(map.data()+was);
	dat->len=len;
	freepos+=alignstart<int>(sizeof(data_t)+len);
	return dat;
	}
void reset() {
	freepos=datastart();
	memset(map.data(),'\0',freepos);
	}
void makenull(int datnr) {
	datpos(datnr)=0;
	}
	
void setpos(int datnr,data_t *dat) {
	int pos=reinterpret_cast<unsigned char *>(dat)-map.data();
	datpos(datnr)=pos;
	}
	};

	class SavedApart {};
string_view getpreviousstate(string_view sbasedir ) ; // delete[] should be called on result
class scanstate: public multimmap {
	class TimeFile {};
	string_view filename;

	data_t *fromfile(const char *filename) ;
	scanstate(TimeFile,string_view datafile):multimmap(4,datafile.data(),4*4096),filename(datafile) {
		LOGGER("scanstate\n");
		}
	public:
//static	string_view getprevious(string_view basedir ) ;
static	string_view makefilename(const string_view basedir,const time_t tim) ;
	virtual ~scanstate() override {
		delete[] filename.data();;
		}
	bool makelink() {
		return makelink(filename);
		}

	static bool makelink(string_view filename);
	scanstate(string_view basedir,time_t tim):scanstate(TimeFile{},makefilename(basedir,tim)) {
		}
	scanstate(int block=1): multimmap(4,block*4096) {}
//	scanstate(string_view basedir) : multimmap(4,getprevious(basedir),4*4096) { }
	scanstate(string_view basedir) : scanstate(TimeFile{}, getpreviousstate(basedir)) { }
	scanstate(string_view prev,SavedApart);
	void removefile(); 
};
struct outobj;

template <class B,class M,class ...Ts>
B * locknew(M &mu,Ts && ... args) {
	LOGGER("locknew\n");
        std::lock_guard<M> lock(mu);
        return new B(std::forward<Ts>(args)...);
        }



class Abbott {
public:
struct scanresult_t {
	const AlgorithmResults *res;
	const ScanData *scan;
    } ;
string serial;
int sensorindex=sensors->sensorindex(serial.data());
SensorGlucoseData *hist=sensors->gethist(sensorindex);

private:
pathconcat sensordir;
data_t *uid;
scanstate *state;
int errorcode=0;
//Sensoren sens;

//Abbott(string_view basedir,string &&sensor,data_t *uidin): serial(std::move(sensor)),sensordir(basedir,serial),uid(uidin),state(hist?(hist->mutex.lock(),new scanstate(sensordir)):nullptr) {
Abbott(string_view basedir,string &&sensor,data_t *uidin): serial(std::move(sensor)),sensordir(basedir,serial),uid(uidin),state(hist?locknew<scanstate>(hist->mutex,sensordir):nullptr) {
//	if(hist) hist->mutex.unlock();
	LOGGER("Abbott(string_view basedir,string &&sensor,data_t *uidin)\n");
	std::filesystem::path dir( sensordir.cbegin(),sensordir.cend());
	 if(!((is_directory(dir)||std::filesystem::create_directories(dir) )&&!access(sensordir.data(), W_OK|R_OK|X_OK))) {
		lerror(sensordir.data());
		errorcode=1;
	 	}
	}
/*
Abbott(string_view basedir,data_t *uidin,int fam=0): sensordir(basedir,getserial(fam,reinterpret_cast<unsigned char *>(uidin->data()))), uid(uidin) {
	init();
	}
	*/
Abbott(string_view basedir,data_t *uidin,int fam=0): Abbott(basedir,getserial(fam,reinterpret_cast<unsigned char *>(uidin->data())), uidin) {
	}
public:
const string_view getsensordir() const {return sensordir;};
const data_t * getsensorid() const {return uid;};
/*
Abbott(string_view basedir,string_view sensor): sensordir(basedir,sensor), uid(data_t::newex(unserial(sensor.data()))) {
	init();
	}
	*/

Abbott(string_view basedir,string_view  sensor): Abbott(basedir,string(sensor), data_t::newex(unserial(sensor.data()))) {
	}

//Abbott(JNIEnv *env,jobject obj,string_view basedir,jbyteArray juid, jbyteArray info);
Abbott(JNIEnv *env,string_view basedir,jbyteArray juid, jbyteArray info);
~Abbott() {
 	data_t::deleteex(uid);
	delete state;
	}
int error() const  {return errorcode;}
scanresult_t ProcessOne(scandata *data)  ;
AlgorithmResults *callAbbottAlg(int startsincebase,scanstate *state,scandata *data,scanstate *newstate,outobj *starttime,outobj *endtime,jobject person) ;
//AlgorithmResults *callAbbottAlg(scanstate *state,scandata *data,scanstate *newstate) ;
scanresult_t callAbbottAlg(scandata *data) ;
scanstate *initnewsensor( scandata *data) ;

};
//AlgorithmResults *callAbbottAlg(ByteArray *uid,scanstate *state,scandata *data,scanstate *newstate);
//extern "C" JNIEXPORT jint JNICALL Java_tk_glucodata_Glucose_init(JNIEnv *env, jclass thiz,jstring dir) ;

//extern scanstate *initnewsensor( scandata *data, ByteArray *uid) ;
//int  abbottinit(std::string_view filesdir,JNIEnv *env=nullptr,jobject thiz=nullptr);
extern int  abbottinit();

bool linklib(const char *filename) ;

typedef int8_t* Bytes_t;

struct timevalues {
int warmup,wear;
};

extern timevalues     patchtimevalues(const data_t *info) ;
extern jbyte  getactivationcommand(const data_t *info) ;
//extern data_t  * getactivationpayload(const data_t * patchid, const data_t * info, jbyte person) ;
data_t  * getactivationpayload(scanstate *state,const data_t * patchid, const data_t * info, jbyte person) ;

extern const ScanData *lastscan;

constexpr time_t basesecs=1262304000L;
constexpr const int endsensorsecs= (14*24-1)*60*60; 
constexpr const int days15=(15*24*60*60);
constexpr const int daysecs=24*60*60;

struct streamdata {
int sensorindex;
SensorGlucoseData *hist;
int startsincebase;
jbyte person=0;
scanstate *state;
#ifndef NORAWSTREAM
int blueuit;
#endif
int index;
public:
//streamdata(const char * sensorname):hist(sensors->gethist(sensorname)), startsincebase(((time(NULL)-hist->getstarttime())>(14*24*60*60-1)?(24*60*60):0)+hist->getstarttime()-basesecs),state((hist->mutex.lock(),new scanstate(hist->getsensordir()))) {
streamdata(const char * sensorname,int index):sensorindex( sensors->sensorindex(sensorname)),hist(sensors->gethist(sensorindex)), startsincebase(((time(NULL)-hist->getstarttime())>=endsensorsecs?daysecs:0)+hist->getstarttime()-basesecs),state(locknew<scanstate>(hist->mutex,hist->getsensordir())),index(index) {
#ifndef NORAWSTREAM
		pathconcat uit(hist->getsensordir(),rawstream);
		blueuit=open(uit.data(),O_APPEND|O_CREAT|O_WRONLY, S_IRUSR |S_IWUSR);
#endif

//		hist->mutex.unlock();

		}

	const data_t *getident() const {
		return hist->getsensorident();
		}
	const data_t *getinfo() const {
		return hist-> getpatchinfo(); 
		}
~streamdata() {
#ifndef NORAWSTREAM
	close(blueuit);
#endif
#ifndef NDEBUG
     delete state;
#endif
	}
bool good() const {
	return state&&state->good();
	}
void setstate(scanstate *newstate) {	
	delete state;
	state=newstate;
	}
void renewstate() {
	scanstate *tmpstate= state;
	state= new scanstate(hist->getsensordir());
	delete tmpstate;
	}

AlgorithmResults *  processTooth(data_t * bluedata, scanstate *newstate,uint32_t nutime,int index) ;
};

extern pathconcat sensorbasedir;
inline  static  const AlgorithmResults  __attribute__((unused)) *Initialized=reinterpret_cast<const AlgorithmResults *>(1ul);
//constexpr static  const AlgorithmResults * Initialized=(const AlgorithmResults *)(1ul);
#endif
